/* This file is part of Zanshin

   Copyright 2015 Sandro Knauß <sknauss@kde.org>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of
   the License or (at your option) version 3 or any later version
   accepted by the membership of KDE e.V. (or its successor approved
   by the membership of KDE e.V.), which shall act as a proxy
   defined in Section 14 of version 3 of the license.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
   USA.
*/

#include "printwidget.h"

#include <QTextDocument>
#include <QTextEdit>
#include <QVBoxLayout>
#include <QTimer>
#include <QEventLoop>
#include <QCoreApplication>

#include <KDE/KLocale>

#include "presentation/artifactfilterproxymodel.h"
#include "presentation/metatypes.h"

#include "akonadi/akonaditaskqueries.h"
#include "akonadi/akonaditaskrepository.h"
#include "presentation/tasklistmodel.h"
#include <presentation/querytreemodelbase.h>

#include "grantlee_paths.h"
#include "grantlee_core.h"

using namespace Widgets;

PrintWidget::PrintWidget(QWidget *parent, ApplicationMode mode)
    : QWidget(parent),
      m_model(0),
      m_mode(mode),
      m_document(new QTextDocument(this)),
      m_edit(new QTextEdit()),
      m_timer(new QTimer()),
      m_repository(new Akonadi::TaskRepository),
      m_queries(new Akonadi::TaskQueries),
      m_engine(0),
      m_init(false)
{
    m_edit->setDocument(m_document);

    connect(m_timer, SIGNAL(timeout()), SLOT(render()));
    m_timer->setInterval(2000);

    auto layout = new QVBoxLayout;
    layout->addWidget(m_edit);
    setLayout(layout);

    QTimer::singleShot(0, this, SLOT(delayedInit()));
}

void PrintWidget::delayedInit()
{
  m_engine = new Grantlee::Engine(this);
  m_engine->addDefaultLibrary( "grantlee_i18n" );
  m_engine->addDefaultLibrary( "grantlee_scriptabletags" );

  Grantlee::FileSystemTemplateLoader::Ptr loader = Grantlee::FileSystemTemplateLoader::Ptr(new Grantlee::FileSystemTemplateLoader());
  loader->setTemplateDirs(QStringList() << GRANTLEE_TEMPLATE_PATH << QCoreApplication::applicationDirPath().replace("bin", "share/apps/zanshin/templates"));
  m_engine->addTemplateLoader(loader);

  m_init = true;
  emit init();
}

QObject *PrintWidget::model() const
{
    return m_model;
}

void PrintWidget::setModel(QObject *model)
{
    if (model == m_model)
        return;

    m_model = model;

    if (!m_model) {
        m_timer->stop();
        return;
    }
    m_timer->start();
}

void PrintWidget::render()
{
    if (!m_model) {
        return;
    }

    auto model = static_cast<Presentation::TaskListModel*>(m_model);

    QString html;

    QVariantList artifacts;

    for(int i=0; i<model->rowCount(); i++) {
        auto index = model->index(i,0);
        auto data = index.data(Presentation::QueryTreeModelBase::ObjectRole);
        if (!data.isValid())
            continue;

        auto artifact = data.value<Domain::Artifact::Ptr>();
        if (!artifact)
            continue;
        artifacts << QVariant::fromValue(static_cast<QObject*>(artifact.data()));
    }

    Grantlee::Template t = m_engine->loadByName("main.html");
    if (t->error()) {
        qWarning() << t->errorString();
        return;
    }
    Grantlee::Context c;
    c.insert("anz", artifacts.count());
    c.insert("tasks", artifacts);

    html += t->render(&c);
    m_document->setHtml(html);
}

void PrintWidget::print(QPrinter *printer)
{
    render();
    m_document->print(printer);
}

void PrintWidget::print(QObject *model, QPrinter *printer)
{
    if (!m_init) {
        QEventLoop loop;
        connect(this, SIGNAL(init()), &loop, SLOT(quit()));
        loop.exec();
    }
    setModel(model);
    print(printer);
}
